//
//  main.cpp
//  ceharness
//
//  Created by Project Zero on 23.09.22.
//

#include <dlfcn.h>
#include <CoreFoundation/CoreFoundation.h>

#define CE_MAX_KEY_SIZE 240

typedef struct {
    const uint8_t* data;
    size_t length;
} CEBuffer;

typedef struct {
    uint8_t data[CE_MAX_KEY_SIZE];
    size_t length;
} CEStaticBuffer;

OS_CLOSED_ENUM(CEQueryOpOpcode, int64_t,
               kCEOpNoop = 0,
               kCEOpSelectKey = 1,
               kCEOpSelectIndex = 2,
               kCEOpMatchString = 3,
               kCEOpMatchStringPrefix = 4,
               kCEOpMatchBool = 5,
               kCEOpStringValueAllowed = 6,
               kCEOpMatchInteger = 7,
               kCEOpStringPrefixValueAllowed = 8,
               kCEOpDynamic = 0x1LL << 62);

typedef struct CEQueryOperation {
    CEQueryOpOpcode_t opcode;
    union {
        CEBuffer dynamicParameter;
        CEStaticBuffer stringParameter;
        int64_t numericParameter;
    } parameters;
} CEQueryOperation_t;

typedef CEQueryOperation_t CEQuery_t[];

#define CEMatchBool(val) (CEQueryOperation_t){.opcode = kCEOpMatchBool, .parameters = {.numericParameter = !!val}}
#define CESelectDictValueDynamic(key, len) (CEQueryOperation_t){.opcode = (CEQueryOpOpcode_t)(kCEOpSelectKey | kCEOpDynamic), .parameters = {.dynamicParameter = {.data = key, .length = len}}}

uint64_t CECRuntime;
uint64_t kCENoError;
uint64_t kCEMalformedEntitlements;

typedef uint64_t (*t_CEManagedContextFromCFData)(uint64_t runtime, CFDataRef data, uint64_t *out_context);
t_CEManagedContextFromCFData CEManagedContextFromCFData;

typedef uint64_t (*t_CEQueryContextToCFDictionary)(uint64_t in_context, CFMutableDictionaryRef* dict);
t_CEQueryContextToCFDictionary CEQueryContextToCFDictionary;

typedef uint64_t (*t_CEAcquireManagedContext)(uint64_t runtime, uint64_t *out_context, uint64_t reserved1, uint64_t reserved2, uint64_t reserved3, uint64_t reserved4, uint64_t type, uint64_t start, uint64_t end);
t_CEAcquireManagedContext CEAcquireManagedContext;

typedef uint64_t (*t_CEAcquireUnmanagedContext)(uint64_t runtime, uint64_t *out_context, uint64_t reserved1, uint64_t reserved2, uint64_t reserved3, uint64_t reserved4, uint64_t type, uint64_t start, uint64_t end);
t_CEAcquireUnmanagedContext CEAcquireUnmanagedContext;

typedef uint64_t (*t_CESizeDeserialization)(uint64_t in_context, uint64_t *num_elements);
t_CESizeDeserialization CESizeDeserialization;

typedef uint64_t (*t_CESizeXMLDeserialization)(char *serialized_data, uint64_t num_elements, uint64_t *out_size);
t_CESizeXMLDeserialization CESizeXMLDeserialization;

typedef uint64_t (*t_CEDeserialize)(uint64_t in_context, char *data, uint64_t num_elements);
t_CEDeserialize CEDeserialize;

typedef uint64_t (*t_CEDeserializeXML)(uint64_t runtime, char *serialized_data, uint64_t num_elements, char *xml_data, size_t xml_size);
t_CEDeserializeXML CEDeserializeXML;

typedef uint64_t (*t_CEContextQuery)(uint64_t in_context, CEQueryOperation_t *query, uint64_t query_length);
t_CEContextQuery CEContextQuery;

void callback_fn() {
  // printf("In callback\n");
  throw 0x13371337;
}

void ResolveSymbols() {
  void *libmis = dlopen("libCoreEntitlements.dylib", RTLD_LAZY);

  CECRuntime = *(uint64_t *)dlsym(libmis, "CECRuntime");
  if(!CECRuntime) {
    printf("Symbol lookup fail\n");
    exit(0);
  }

  kCENoError = *(uint64_t *)dlsym(libmis, "kCENoError");
  if(!kCENoError) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  kCEMalformedEntitlements = *(uint64_t *)dlsym(libmis, "kCEMalformedEntitlements");
  if(!kCEMalformedEntitlements) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  

  CEManagedContextFromCFData = (t_CEManagedContextFromCFData)dlsym(libmis, "CEManagedContextFromCFData");
  if(!CEManagedContextFromCFData) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CEQueryContextToCFDictionary = (t_CEQueryContextToCFDictionary)dlsym(libmis, "CEQueryContextToCFDictionary");
  if(!CEQueryContextToCFDictionary) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CEAcquireManagedContext = (t_CEAcquireManagedContext)dlsym(libmis, "CEAcquireManagedContext");
  if(!CEAcquireManagedContext) {
    printf("Symbol lookup fail\n");
    exit(0);
  }

  CEAcquireUnmanagedContext = (t_CEAcquireUnmanagedContext)dlsym(libmis, "CEAcquireUnmanagedContext");
  if(!CEAcquireUnmanagedContext) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CESizeDeserialization = (t_CESizeDeserialization)dlsym(libmis, "CESizeDeserialization");
  if(!CESizeDeserialization) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CESizeXMLDeserialization = (t_CESizeXMLDeserialization)dlsym(libmis, "CESizeXMLSerialization");
  if(!CESizeXMLDeserialization) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CEDeserialize = (t_CEDeserialize)dlsym(libmis, "CEDeserialize");
  if(!CEDeserialize) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CEDeserializeXML = (t_CEDeserializeXML)dlsym(libmis, "CESerializeXML");
  if(!CEDeserializeXML) {
    printf("Symbol lookup fail\n");
    exit(0);
  }
  
  CEContextQuery = (t_CEContextQuery)dlsym(libmis, "CEContextQuery");
  if(!CEContextQuery) {
    printf("Symbol lookup fail\n");
    exit(0);
  }

  *((uint64_t *)(CECRuntime+32)) = (uint64_t)(&callback_fn);
}

void parse(char *filename) {
  char *sample_bytes = NULL;
  uint32_t sample_size = 0;
  
  FILE *fp = fopen(filename, "rb");
  if(!fp) {
    printf("Error opening %s\n", filename);
    return;
  }
  fseek(fp, 0, SEEK_END);
  sample_size = ftell(fp);
  fseek(fp, 0, SEEK_SET);
  sample_bytes = (char *)malloc(sample_size);
  fread(sample_bytes, 1, sample_size, fp);
  fclose(fp);

  char entitlement[] = "task_for_pid-allow";
  CEQuery_t query = {
    CESelectDictValueDynamic((const uint8_t*)entitlement, sizeof(entitlement) - 1),
    CEMatchBool(true)
  };
    
  uint64_t *context = 0;
  CFDataRef data = CFDataCreate(kCFAllocatorDefault, (const unsigned char *)sample_bytes, sample_size);
  uint64_t ret = (*CEManagedContextFromCFData)(CECRuntime, data, (uint64_t *)(&context));
  if(ret == kCENoError) {
    printf("Validation successful\n");
    CFMutableDictionaryRef ceDict;
    ret = (*CEQueryContextToCFDictionary)((uint64_t)context, &ceDict);
    if(ret == kCENoError) {
      printf("Successfully converted entitlements to dictionary\n");
      printf("Dictionary:\n%s\n", CFStringGetCStringPtr(CFCopyDescription(ceDict), kCFStringEncodingASCII));
    } else {
      printf("Error converting entitlements to dictionary\n");
    }
    printf("Querying if entitlements contain task_for_pid-allow\n");
    ret = (*CEContextQuery)((uint64_t)context, query, 2);
    if(ret == kCENoError) {
      printf("Query said yes\n");
    } else {
      printf("Query said no\n");
    }
  } else {
    printf("Validation failed\n");
  }
  
  if(sample_bytes) free(sample_bytes);
}


int main(int argc, char * argv[]) {
  if(argc < 2) {
    printf("Usage: %s <file>\n", argv[0]);
    return 0;
  }

  ResolveSymbols();
  
  parse(argv[1]);

  return 0;
}
